
// Copyright in this code is held by Dr B. R. Gladman but free direct or
// derivative use is permitted subject to acknowledgement of its origin.  
//
// Dr. B. R. Gladman (gladman@seven77.demon.co.uk).
//
// AES Algorithm Test - Generate KAT and Monte Carlo test results and 
// compare the answers with the reference test vectors in a standard 
// set of test files
//
// usage: aes_gav /a:[12345] /t:[kec] [/h]  where:
//  /a:[12345] algorithms to be included in the test
//  /t:[kec]   type of test
//  /h         use author's byte order (serpent)
// where:
//  1:mars, 2:rc6, 3:rijndael, 4:serpent, 5:twofish
//  k: run ECB Known Answer tests  
//  e: run ECB Monte Carlo tests
//  c: run CBC Monte Carlo tests
//
// Note that, when compared with the NIST test vector sequences, this code 
// implements one additional (all zero) test vector as the first vector in 
// each set (test 0).

#include <iostream>
#include <fstream>
#include <iomanip>
#include <cctype>
#include <string>

#include "aes_config.h"
#include "aes_defs.h"
#include "aes_aux.h"
#include "mars.h"
#include "rc6.h"
#include "rijndael.h"
#include "serpent.h"
#include "twofish.h"

// Serpent hack to get test agreement because the test vectors for this 
// algorithm assemble the key vectors in the reverse direction in memory

bool    serpent_hack = false;

enum test_type { ecb_vk, ecb_vt, ecb_me, ecb_md, cbc_me, cbc_md };

void ref_test(const std::string &in_file, const int it_cnt, test_type t_type, AES& alg)
{   u4byte			i, kl, test_no, cnt, e_cnt;
    u1byte			key[32], pt[16], iv[16], ect[16], act[32];
	std::ifstream	inf;
	std::string		str;
    int				ty;

    std::cout << std::endl << "Test file: " << in_file << std::endl << "Status: ";

    inf.open(in_file.c_str(), std::ios::in);	// reference file for test vectors

    if(!inf.is_open())	// if file is not present
    {
        std::cout << "error in running test" << std::endl; return;
    }

    cnt = 0; e_cnt = test_no = 0;

    while(!inf.eof())				// while there are tests
    {
        ty = find_line(inf, str);	// input a line

        if(ty < 0)					// until end of file

            continue;

        switch(ty)		// process line type
        {
          case 0:   kl = get_dec(str.begin() + 8); continue;			// key length
          case 1:   test_no = get_dec(str.begin() + 2); continue;		// test number
          case 2:   block_in(iv, str.begin()+ 3, str.end()); continue;	// init vector
          case 3:   block_in(key, str.begin()+ 4, str.end()); continue;	// key
          case 4:   block_in(pt, str.begin()+ 3, str.end());			// plaintext
                    if(t_type != ecb_md && t_type != cbc_md)
                        continue;
                    break;
          case 5:   block_in(ect, str.begin()+ 3, str.end());			// ciphertext
                    if(t_type == ecb_md || t_type == cbc_md)
                        continue;
                    break;
        }

		if(serpent_hack)
		
			block_reverse(key, kl / 8);

        alg.set_key(key, kl);		// set the key

#if KEEP_USER_HAPPY
        if(it_cnt > 100)			// output during heavy processing
		{							// to keep user happy
			if(!(test_no % 40))

				std::cout << std::endl;

			std::cout << '.';
		}
#endif
        if(t_type == ecb_md || t_type == cbc_md)
        {
            block_copy(act, ect);				// encrypted text to low block

            if(t_type == cbc_md)				// CBC Monte Carlo decryption
            {
                block_copy(act + 16, iv);		// IV to high block

                for(i = 0; i < it_cnt; i += 2)	// do decryptions two at a time
                {
					if(serpent_hack)
				
						block_reverse(act);

                    alg.decrypt(act, ect);		// decrypt low block

					if(serpent_hack)
					{
						block_reverse(act); block_reverse(ect);

					}

					block_xor(act + 16, ect);	// xor into high block

					if(serpent_hack)
				
						block_reverse(act + 16);

                    alg.decrypt(act + 16, ect);	// decrypt high block

					if(serpent_hack)
					{				
						block_reverse(ect); block_reverse(act + 16);
					}

					block_xor(act, ect);		// xor into low block
                }
            }
            else	// ECB Monte Carlo decryption 
            {
				if(serpent_hack)

					block_reverse(act);

                for(i = 0; i < it_cnt; ++i)
        
                    alg.decrypt(act, act);
		
				if(serpent_hack)

					block_reverse(act);
			}

            if(!block_cmp(pt, act))
            {
                std::cout << std::endl << "decryption error on test " << test_no; e_cnt++;
            }

            if(t_type == ecb_md)	// test encryption if ECB mode
            {
				if(serpent_hack)

					block_reverse(act);

                for(i = 0; i < it_cnt; ++i)

                    alg.encrypt(act, act); 

				if(serpent_hack)

					block_reverse(act);

                if(!block_cmp(ect, act))
                {   
                    std::cout << std::endl << "encryption error on test " << test_no; e_cnt++;
                }
            }
        }
        else	// if(t_type == ecb_me || t_type == cbc_me)
        {
            if(t_type == cbc_me)						// CBC Monte Carlo encryption
            {
                block_copy(act, iv); block_copy(act + 16, pt);	// copy IV and plaintext

                for(i = 0; i < it_cnt; i += 2)
                {
                    block_xor(act + 16, act);			// xor low block into high block

					if(serpent_hack)

						block_reverse(act + 16);

                    alg.encrypt(act + 16, act + 16);	// encrypt high block

					if(serpent_hack)

						block_reverse(act + 16);

                    block_xor(act, act + 16);			// xor high block into low block

					if(serpent_hack)

						block_reverse(act);

                    alg.encrypt(act, act);				// encrypt low block
		
					if(serpent_hack)

						block_reverse(act);
                }
            }
            else	// ECB Monte Carlo encryption
            {
                block_copy(act, pt);

				if(serpent_hack)

					block_reverse(act);

                for(i = 0; i < it_cnt; ++i)
        
                    alg.encrypt(act, act);

				if(serpent_hack)
				
					block_reverse(act);
            }

            if(!block_cmp(ect, act))
            {
                std::cout << std::endl << "encryption error on test " << test_no; e_cnt++;
            }
        
            if(t_type == ecb_me)	// if ECB mode test decrytpion
            {
				if(serpent_hack)
				
					block_reverse(act);

                for(i = 0; i < it_cnt; ++i)

                    alg.decrypt(act, act); 

				if(serpent_hack)
				
					block_reverse(act);

                if(!block_cmp(pt, act))
                {   
                    std::cout << std::endl << "decryption error on test " << test_no; e_cnt++;
                }
            }
        }
    }

    inf.close();

    if(e_cnt > 0)	// report any errors

        std::cout << std::endl << e_cnt << " errors during test" << std::endl;

    else			// else report all is well

        std::cout << "all tests correct" << std::endl;
}

void do_tests(const bool vkt, const bool ecb, const bool cbc, AES& alg)
{   std::string		name, path;

    std::cout << std::endl << "Run tests for the " << alg.name() << " algorithm (" 
                    << alg.name() << ".cpp)" << std::endl;

    path = ref_path + std::string(alg.name());
    
    if(vkt)
    {
        name = path + aes_name[0]; ref_test(name, 1, ecb_vk, alg);
        name = path + aes_name[1]; ref_test(name, 1, ecb_vt, alg);
    }

    if(ecb)
    {
        name = path + aes_name[2];  ref_test(name, 10000, ecb_me, alg);
        name = path + aes_name[3];  ref_test(name, 10000, ecb_md, alg);
    }

    if(cbc)
    {
        name = path + aes_name[4];  ref_test(name, 10000, cbc_me, alg);
        name = path + aes_name[5];  ref_test(name, 10000, cbc_md, alg);
    }

    std::cout << std::endl;
}

void main(int argc, char *argv[])
{	bool		vkt, ecb, cbc;
	mars		_mars;
	rc6			_rc6;
	rijndael	_rijndael;
	serpent		_serpent;
	twofish		_twofish;

    if(argc == 1)
    {
        std::cout << std::endl << "usage: aes_rav /a:[12345] /t:[kec] [/h]  where:";
        std::cout << std::endl << "  /a:[12345] algorithms to be tested";
        std::cout << std::endl << "  /t:[kec]   type of test";
        std::cout << std::endl << "  /h         use author's byte order (serpent)";
        std::cout << std::endl << "where:";
        std::cout << std::endl << "  1:mars, 2:rc6, 3:rijndael, 4:serpent, 5:twofish";
        std::cout << std::endl << "  k: run ECB Known Answer tests";  
        std::cout << std::endl << "  e: run ECB Monte Carlo tests";
        std::cout << std::endl << "  c: run CBC Monte Carlo tests";
        std::cout << std::endl << std::endl; exit(0);
    }

    vkt = test_args(argc, argv, 't', 'k');
    ecb = test_args(argc, argv, 't', 'e');
    cbc = test_args(argc, argv, 't', 'c');
	
	serpent_hack = false;

	if(test_args(argc, argv, 'a', '1'))

		do_tests(vkt, ecb, cbc, _mars);
	
	if(test_args(argc, argv, 'a', '2'))

		do_tests(vkt, ecb, cbc, _rc6);
	
	if(test_args(argc, argv, 'a', '3'))

		do_tests(vkt, ecb, cbc, _rijndael);
	
	if(test_args(argc, argv, 'a', '4'))
	{
		serpent_hack = test_args(argc, argv, 'h');
		do_tests(vkt, ecb, cbc, _serpent); serpent_hack = false;
	}

	if(test_args(argc, argv, 'a', '5'))

		do_tests(vkt, ecb, cbc, _twofish);
}
