
//  Measure the Encryption, Decryption and Key Setup Times for 
//  an AES Algorithms Using the Pentium Time Stamp Counter

//  Copyright Dr Brian Gladman (gladman@seven77.demon.co.uk)
//	August 1999.  Permission for free use is hereby given
//	provided that the origin of this code is acknowledged. 
//  
//  usage: aes_tmr /a:[12345] where:
//    /a:[12345] algorithms to be tested
//  where:
//    1:mars, 2:rc6, 3:rijndael, 4:serpent, 5:twofish

#include <iostream>
#include <iomanip>

#include "aes_defs.h"
#include "mars.h"
#include "rc6.h"
#include "rijndael.h"
#include "serpent.h"
#include "twofish.h"

u1byte rand8(void);			// pseudo random number generator

const unsigned int loops = 100;	// number of timing loops

double cycles(void)				// read the Pentium Time Stamp Counter
{   u4byte  hi,lo;

    __asm   
    {
        _emit   0x0f			// complete pending operations
        _emit   0xa2
        _emit   0x0f			// read time stamp counter
        _emit   0x31
        mov     lo,eax
        mov     hi,edx
        _emit   0x0f			// complete pending operations
        _emit   0xa2
    }

    return 4294967296.0 * hi + lo;
}

// fill a block with random charactrers

void rand_fill(u1byte *p, u4byte no = 16)
{
	for(int i = 0; i < no; ++i)

		p[i] = rand8();
}

// measure cycles for an encryption call

double e_cycles(const u4byte key_len, AES& alg)
{   u1byte      pt[16], ct[16], key[32];
    double      cy0, cy1, cy2, c1, c2;

    // set up a random key of 256 bits

	rand_fill(key, 32);

    // set up a random plain text

	rand_fill(pt);

    // do a set_key in case it is necessary

    alg.set_key(key, key_len); c1 = c2 = 1.0e100;

    // do an encrypt to remove any 'first time through' effects

    alg.encrypt(pt, ct);

    for(int i = 0; i < loops; ++i)
    {
        rand_fill(pt);
		// time one and two encryptions

        cy0 = cycles();
	    alg.encrypt(pt, ct);
        cy1 = cycles();
	    alg.encrypt(pt, ct);
	    alg.encrypt(pt, ct);
        cy2 = cycles();

        cy2 -= cy1; cy1 -= cy0;		// time for one and two calls

        c1 = (c1 > cy1 ? cy1 : c1);	// find minimum values over the loops

        c2 = (c2 > cy2 ? cy2 : c2);
    }

    return c2 - c1;	// return one call timing
}

// measure cycles for a decryption call

double d_cycles(const u4byte key_len, AES& alg)
{   u1byte      pt[16], ct[16], key[32];
    double      cy0, cy1, cy2, c1, c2;

    // set up a random key of 256 bits

	rand_fill(key, 32);

    // set up a random plain text

	rand_fill(pt);

    // do a set_key in case it is necessary

    alg.set_key(key, key_len); c1 = c2 = 1.0e100;

    // do an decrypt to remove any 'first time through' effects

    alg.decrypt(pt, ct);
 
    for(int i = 0; i < loops; ++i)
    {
		rand_fill(pt);

		// time one and two encryptions

        cy0 = cycles();
	    alg.decrypt(pt, ct);
        cy1 = cycles();
	    alg.decrypt(pt, ct);
	    alg.decrypt(pt, ct);
        cy2 = cycles();

        cy2 -= cy1; cy1 -= cy0;		// time for one and two calls

        c1 = (c1 > cy1 ? cy1 : c1);	// find minimum values over the loops

        c2 = (c2 > cy2 ? cy2 : c2);
    }

    return c2 - c1;	// return one call timing
}

// measure cycles for a key setup

#pragma optimize("", off)		// VC++ optimise has problems here 

double k_cycles(const u4byte key_len, AES& alg)
{   u1byte      key[32];
    double      cy0, cy1, cy2, c1, c2;

    // set up a random key of 256 bits

	rand_fill(key, 32);

    // do an set_key to remove any 'first time through' effects

    alg.set_key(key, key_len); c1 = c2 = 1.0e100;

    for(int i = 0; i < loops; ++i)
    {
		rand_fill(key, 32);

		// time one and two encryptions

        cy0 = cycles();
        alg.set_key(key, key_len);
        cy1 = cycles();
        alg.set_key(key, key_len);
        alg.set_key(key, key_len);
        cy2 = cycles();

        cy2 -= cy1; cy1 -= cy0;		// time for one and two calls

        c1 = (c1 > cy1 ? cy1 : c1);	// find minimum values over the loops

        c2 = (c2 > cy2 ? cy2 : c2);
    }

    return c2 - c1;	// return one call timing
}

#pragma optimize("", on)

void output(const u4byte klen, const double kt, const double et, const double dt)
{   double  av;

    std::cout << std::endl << "//  " << std::setw(3) << klen << " bit key:";

    std::cout << std::endl << "//  Key Setup:" 
			  <<  std::setprecision(0) << std::setw(8) << kt << " cycles";

    std::cout << std::endl << "//  Encrypt:  " 
			  << std::setprecision(0) << std::setw(8) << et << " cycles ="
			  << std::setprecision(1) << std::setw(8) << 25600 / et << " mbits/sec";

    std::cout << std::endl << "//  Decrypt:  " 
		      << std::setprecision(0) << std::setw(8) << dt << " cycles ="
              << std::setprecision(1) << std::setw(8) << 25600 / dt << " mbits/sec";

    av = 0.5 * (et + dt);
    std::cout << std::endl << "//  Mean:     " 
		      << std::setprecision(0) << std::setw(8) << av << " cycles ="
              << std::setprecision(1) << std::setw(8) << 25600 / av << " mbits/sec";
}

// Input command line arguments

bool test_args(int argc, char *argv[], char des_chr, char tst_chr = '\0')
{
    for(int i = 1; i < argc; ++i)
    {
        if(argv[i][0] != '-' && argv[i][0] != '/')

            continue;

        if(argv[i][1] != tolower(des_chr) && argv[i][1] != toupper(des_chr))
		
			continue;

		else if(!tst_chr)

			return true;

        if(argv[i][2] != ':')

            continue;

        for(int j = 3; argv[i][j]; ++j)

			if(argv[i][j] == tolower(tst_chr) || argv[i][j] == toupper(tst_chr))

				return true;
    }

    return false;
}

void time(AES& alg)
{	double		kt, et, dt;
    u4byte		kl;

    std::cout << "//  Algorithm " << alg.name() << " (" << alg.name() << ".cpp)";

    std::cout.setf(std::ios::floatfield, std::ios::fixed);

    for(kl = 128; kl <= 256; kl += 64)
    {
        kt = k_cycles(kl, alg); 
        dt = d_cycles(kl, alg); 
        et = e_cycles(kl, alg); 
        
        output(kl, kt, et, dt);
    }

    std::cout << std::endl << std::endl;
}

void main(int argc, char *argv[])
{	mars		_mars;
	rc6			_rc6;
	rijndael	_rijndael;
	serpent		_serpent;
	twofish		_twofish;

    if(argc == 1)
    {
        std::cout << std::endl << "usage: aes_tmr /a:[12345] where:";
        std::cout << std::endl << "  /a:[12345] algorithms to be tested";
        std::cout << std::endl << "where:";
        std::cout << std::endl << "  1:mars, 2:rc6, 3:rijndael, 4:serpent, 5:twofish";
        std::cout << std::endl << std::endl; exit(0);
    }

	if(test_args(argc, argv, 'a', '1'))

		time(_mars);
	
	if(test_args(argc, argv, 'a', '2'))

		time(_rc6);
	
	if(test_args(argc, argv, 'a', '3'))

		time(_rijndael);

	if(test_args(argc, argv, 'a', '4'))

		time(_serpent);

	if(test_args(argc, argv, 'a', '5'))

		time(_twofish);
}
